#include <unistd.h>
#include <string.h>
#include <pwd.h>
#include <syslog.h>
#include <security/pam_userpass.h>
#include <fcntl.h>
#include <stdio.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef PROGNAME
#define PROGNAME "chkpwd-pam"
#endif
#define LOCKDIR "/var/lock/" PROGNAME
#define THROTTLE 3

typedef enum {
	AuthOK = 0,
	AuthBad = 1,
	AuthError = 2,
	AuthAbort = 3
} AuthReturn;

#if THROTTLE
static int throttle(uid_t uid)
{
	if (!chdir(LOCKDIR)) {
		int fd;
		char fname[sizeof(uid) * 2 + 1];

		sprintf(fname, "%0*X", (int)(sizeof(fname) - 1), uid);
		fd = open(fname, O_RDWR|O_CREAT|O_EXCL|O_NOFOLLOW, 0600);
		if (fd >= 0)
			return close(fd);
		if (errno == EEXIST) {
			fd = open(fname, O_RDWR|O_NOFOLLOW);
			if (fd >= 0) {
				struct stat st;
				struct flock lk = {
					.l_type = F_WRLCK,
					.l_whence = SEEK_SET,
					.l_start = 0,
					.l_len = 0
				};

				if (!fcntl(fd, F_SETLKW, &lk) && !fstat(fd, &st)) {
					time_t t = st.st_mtime + THROTTLE - time(0);

					if (t > 0) {
						if (t > THROTTLE)
							t = THROTTLE;
						do {
							t = sleep(t);
						} while (t);
					}

					if (!futimens(fd, NULL))
						return close(fd);
				}
				close(fd);
			}
		}
	}
	return -1;
}
#endif

int main(void)
{
	char password[256];
	pam_userpass_t data;
	pam_handle_t *h;
	ssize_t l;
	int err;
	struct passwd *pwent;
	struct pam_conv conv = {.conv = pam_userpass_conv, .appdata_ptr = &data};
	uid_t uid = getuid();

#if THROTTLE
	if (throttle(uid))
		return AuthError;
#endif

	pwent = getpwuid(uid);
	if (!pwent || !pwent->pw_name)
		return AuthError;
	data.user = pwent->pw_name;
	l = read(STDIN_FILENO, password, sizeof(password) - 1);
	if (l < 0)
		return AuthError;
	password[l] = '\0';
	data.pass = password;

	openlog(PROGNAME, LOG_PID, LOG_AUTH);

	if (pam_start("userpass", data.user, &conv, &h) != PAM_SUCCESS)
		return AuthError;

	err = pam_authenticate(h, 0);
	pam_end(h, err);
	memset(password, 0, l);

	switch (err) {
	case PAM_SUCCESS:
		return AuthOK;
	case PAM_ABORT:
		return AuthAbort;
	case PAM_USER_UNKNOWN:
	case PAM_AUTH_ERR:
	case PAM_MAXTRIES:
	case PAM_AUTHINFO_UNAVAIL:
		return AuthBad;
	case PAM_CRED_INSUFFICIENT:
	default:
		return AuthError;
	}
}
